from unittest import TestCase

from bigquery.job.config.ems_load_job_config import EmsLoadJobConfig

from bigquery.job.config.ems_job_config import EmsJobPriority, EmsCreateDisposition, EmsWriteDisposition

SCHEMA = {"fields": [{"type": "INT64", "name": "f"}]}


class TestEmsLoadJobConfig(TestCase):
    def setUp(self):
        self.ems_load_job_config = EmsLoadJobConfig(destination_project_id="test_project",
                                                    destination_dataset="test_dataset",
                                                    destination_table="test_table",
                                                    create_disposition=EmsCreateDisposition.CREATE_IF_NEEDED,
                                                    write_disposition=EmsWriteDisposition.WRITE_APPEND,
                                                    schema=SCHEMA)

    def test_destination_project_id(self):
        self.assertEqual(self.ems_load_job_config.destination_project_id, "test_project")

    def test_destination_dataset(self):
        self.assertEqual(self.ems_load_job_config.destination_dataset, "test_dataset")

    def test_create_disposition(self):
        self.assertEqual(self.ems_load_job_config.create_disposition, EmsCreateDisposition.CREATE_IF_NEEDED)

    def test_write_disposition(self):
        self.assertEqual(self.ems_load_job_config.write_disposition, EmsWriteDisposition.WRITE_APPEND)

    def test_destination_table(self):
        self.assertEqual(self.ems_load_job_config.destination_table, "test_table")

    def test_destination_project_id_ifProjectIdIsNone_raisesValueError(self):
        load_config = EmsLoadJobConfig(destination_project_id=None, schema=SCHEMA)

        with self.assertRaises(ValueError):
            load_config.destination_project_id

    def test_destination_project_id_ifProjectIdIsEmptyString_raisesValueError(self):
        load_config = EmsLoadJobConfig(destination_project_id="", schema=SCHEMA)

        with self.assertRaises(ValueError):
            load_config.destination_project_id

    def test_destination_project_id_ifProjectIdIsMultipleWhitespaces_raisesValueError(self):
        load_config = EmsLoadJobConfig(destination_project_id="     \t  ", schema=SCHEMA)

        with self.assertRaises(ValueError):
            load_config.destination_project_id

    def test_destination_dataset_ifDatasetIsNone_raisesValueError(self):
        load_config = EmsLoadJobConfig(destination_dataset=None, schema=SCHEMA)

        with self.assertRaises(ValueError):
            load_config.destination_dataset

    def test_destination_dataset_ifDatasetIsEmptyString_raisesValueError(self):
        load_config = EmsLoadJobConfig(destination_dataset="", schema=SCHEMA)

        with self.assertRaises(ValueError):
            load_config.destination_dataset

    def test_destination_table_ifTableIsNone_raisesValueError(self):
        load_config = EmsLoadJobConfig(destination_table=None, schema=SCHEMA)

        with self.assertRaises(ValueError):
            load_config.destination_table

    def test_destination_table_ifTableIsEmptyString_raisesValueError(self):
        load_config = EmsLoadJobConfig(destination_table="", schema=SCHEMA)

        with self.assertRaises(ValueError):
            load_config.destination_table



from concurrent.futures import Future

from google.cloud.pubsub_v1 import PublisherClient


class EmsPublisherClient:
    __client = PublisherClient()

    def publish(self, topic: str, data: bytes, **attrs) -> Future:
        return self.__client.publish(topic=topic, data=data, **attrs)


import logging
from collections import Iterable
from datetime import datetime

from google.cloud.bigquery.schema import _parse_schema_resource

from bigquery.job.config.ems_job_config import EmsJobPriority
from google.api_core.exceptions import GoogleAPIError
from google.cloud import bigquery
from google.cloud.bigquery import QueryJobConfig, QueryJob, TableReference, DatasetReference, TimePartitioning, \
    LoadJobConfig

from bigquery.ems_api_error import EmsApiError
from bigquery.job.config.ems_load_job_config import EmsLoadJobConfig
from bigquery.job.config.ems_query_job_config import EmsQueryJobConfig
from bigquery.job.ems_job_state import EmsJobState
from bigquery.job.ems_query_job import EmsQueryJob

logger = logging.getLogger(__name__)

RETRY = "-retry-"


class EmsBigqueryClient:
    def __init__(self, project_id: str, location: str = "EU"):
        self.__project_id = project_id
        self.__bigquery_client = bigquery.Client(project_id)
        self.__location = location

    def get_job_list(self, min_creation_time: datetime = None, max_result: int = 20) -> Iterable:
        """
        Args:
            min_creation_time (datetime.datetime, optional):
                If set, only jobs created after or at this timestamp are returned.
                If the datetime has no time zone assumes UTC time.
            max_result (int, optional):
                Maximum number of jobs to return.
        Yields:
            EmsQueryJob: the next job
        """
        for job in self.__bigquery_client.list_jobs(all_users=True,
                                                    max_results=max_result,
                                                    min_creation_time=min_creation_time):
            if isinstance(job, QueryJob):
                destination = job.destination
                table_id, dataset_id, project_id = \
                    (destination.table_id, destination.dataset_id, destination.project) \
                    if destination is not None else (None, None, None)

                config = EmsQueryJobConfig(priority=EmsJobPriority[job.priority],
                                           destination_project_id=project_id,
                                           destination_dataset=dataset_id,
                                           destination_table=table_id,
                                           create_disposition=job.create_disposition,
                                           write_disposition=job.write_disposition)
                yield EmsQueryJob(job.job_id, job.query,
                                  config,
                                  EmsJobState(job.state),
                                  job.error_result)

    def get_jobs_with_prefix(self, job_prefix: str, min_creation_time: datetime, max_result: int = 20) -> list:
        jobs = self.get_job_list(min_creation_time, max_result)
        matched_jobs = filter(lambda x: job_prefix in x.job_id, jobs)
        return list(matched_jobs)

    def relaunch_failed_jobs(self, job_prefix: str, min_creation_time: datetime, max_attempts: int = 3) -> list:
        def launch(job: EmsQueryJob) -> str:
            prefix_with_retry = self.__decorate_id_with_retry(job.job_id, job_prefix, max_attempts)
            return self.run_async_query(job.query, prefix_with_retry, job.query_config)

        jobs = self.get_jobs_with_prefix(job_prefix, min_creation_time)
        failed_jobs = [x for x in jobs if x.is_failed]
        return [launch(job) for job in failed_jobs]

    def run_async_query(self,
                        query: str,
                        job_id_prefix: str = None,
                        ems_query_job_config: EmsQueryJobConfig = EmsQueryJobConfig(
                            priority=EmsJobPriority.INTERACTIVE)) -> str:
        return self.__execute_query_job(query=query,
                                        ems_query_job_config=ems_query_job_config,
                                        job_id_prefix=job_id_prefix).job_id

    def run_async_load_job(self, source_uri: str, job_id_prefix: str, config: EmsLoadJobConfig) -> str:
        # TODO wrap into own error (do not let google exception to leak)
        return self.__bigquery_client.load_table_from_uri(source_uris=source_uri,
                                                          destination=TableReference(
                                                              DatasetReference(config.destination_project_id,
                                                                               config.destination_dataset),
                                                              config.destination_table),
                                                          job_id_prefix=job_id_prefix,
                                                          location=self.__location,
                                                          job_config=self.__create_load_job_config(config)).job_id

    def run_sync_query(self,
                       query: str,
                       ems_query_job_config: EmsQueryJobConfig = EmsQueryJobConfig(priority=EmsJobPriority.INTERACTIVE)
                       ) -> Iterable:
        logger.info("Sync query executed with priority: %s", ems_query_job_config.priority)
        try:
            return self.__get_mapped_iterator(
                self.__execute_query_job(query=query,
                                         ems_query_job_config=ems_query_job_config).result()
            )
        except GoogleAPIError as e:
            raise EmsApiError("Error caused while running query | {} |: {}!".format(query, e.args[0]))

    def __decorate_id_with_retry(self, job_id: str, job_prefix: str, retry_limit: int):
        retry_counter = 0
        if RETRY in job_id:
            retry_counter = self.__get_retry_counter(job_id, job_prefix)
        prefix_with_retry = job_prefix + RETRY + str(retry_counter + 1) + "-"

        if retry_counter >= retry_limit - 1:
            raise RetryLimitExceededError()
        return prefix_with_retry

    def __get_retry_counter(self, job_id, job_id_prefix):
        return int(job_id[len(job_id_prefix) + len(RETRY)])

    def __execute_query_job(self, query: str, ems_query_job_config: EmsQueryJobConfig, job_id_prefix=None) -> QueryJob:
        return self.__bigquery_client.query(query=query,
                                            job_config=(self.__create_job_config(ems_query_job_config)),
                                            job_id_prefix=job_id_prefix,
                                            location=self.__location)

    def __create_load_job_config(self, ems_load_job_config: EmsLoadJobConfig) -> LoadJobConfig:
        config = LoadJobConfig()
        config.create_disposition = ems_load_job_config.create_disposition.value
        config.write_disposition = ems_load_job_config.write_disposition.value
        config.schema = _parse_schema_resource(ems_load_job_config.schema)
        return config

    def __create_job_config(self, ems_query_job_config: EmsQueryJobConfig) -> QueryJobConfig:
        job_config = QueryJobConfig()
        job_config.priority = ems_query_job_config.priority.value
        job_config.use_legacy_sql = False
        job_config.use_query_cache = ems_query_job_config.use_query_cache
        if ems_query_job_config.destination_table is not None:
            job_config.time_partitioning = TimePartitioning("DAY")
            table_reference = TableReference(
                DatasetReference(ems_query_job_config.destination_project_id or self.__project_id,
                                 ems_query_job_config.destination_dataset),
                ems_query_job_config.destination_table)
            job_config.destination = table_reference
            job_config.write_disposition = ems_query_job_config.write_disposition.value
            job_config.create_disposition = ems_query_job_config.create_disposition.value
        return job_config

    @staticmethod
    def __get_mapped_iterator(result: Iterable):
        for row in result:
            yield dict(list(row.items()))


class RetryLimitExceededError(Exception):
    pass



from typing import Union

from bigquery.job.config.ems_load_job_config import EmsLoadJobConfig
from bigquery.job.ems_job_state import EmsJobState


class EmsLoadJob:
    def __init__(self,
                 job_id: str,
                 load_config: EmsLoadJobConfig,
                 state: EmsJobState,
                 error_result: Union[dict, None]):
        self.__job_id = job_id
        self.__load_config = load_config
        self.__state = state
        self.__error_result = error_result

    @property
    def load_config(self) -> EmsLoadJobConfig:
        return self.__load_config

    @property
    def state(self) -> EmsJobState:
        return self.__state

    @property
    def job_id(self) -> str:
        return self.__job_id

    @property
    def is_failed(self) -> bool:
        return self.__error_result is not None



from typing import Callable

from google.cloud.pubsub_v1 import SubscriberClient
from google.cloud.pubsub_v1.subscriber.message import Message

from pubsub.ems_streaming_future import EmsStreamingFuture
from pubsub.ems_message import EmsMessage


class EmsSubscriberClient:
    __client = SubscriberClient()

    def subscribe(self, subscription: str, callback: Callable[[EmsMessage], None]) -> EmsStreamingFuture:
        def callback_wrapper(message: Message) -> None:
            callback(EmsMessage(message.data))
            message.ack()

        future = self.__client.subscribe(
            subscription=subscription,
            callback=callback_wrapper
        )

        return EmsStreamingFuture(future)



from unittest import TestCase

from bigquery.job.config.ems_load_job_config import EmsLoadJobConfig
from bigquery.job.ems_job_state import EmsJobState
from bigquery.job.ems_load_job import EmsLoadJob


class TestEmsLoadJob(TestCase):

    def setUp(self):
        self.load_config = EmsLoadJobConfig(schema={"fields": [{"type": "INT64", "name": "f"}]})
        error_result = {"some": "error", "happened": "here"}
        self.ems_load_job = EmsLoadJob("test-job-id", self.load_config, EmsJobState.DONE, error_result)

    def test_state(self):
        self.assertEqual(self.ems_load_job.state, EmsJobState.DONE)

    def test_job_id(self):
        self.assertEqual(self.ems_load_job.job_id, "test-job-id")

    def test_is_failed(self):
        self.assertTrue(self.ems_load_job.is_failed)

    def test_is_not_failed(self):
        not_failed_ems_load_job = EmsLoadJob("test-job-id", self.load_config, EmsJobState.DONE, None)

        self.assertFalse(not_failed_ems_load_job.is_failed)



from bigquery.job.config.ems_job_config import EmsJobConfig, EmsJobPriority


class EmsQueryJobConfig(EmsJobConfig):

    def __init__(self,
                 priority: EmsJobPriority = EmsJobPriority.INTERACTIVE,
                 use_query_cache: bool = False,
                 *args, **kwargs):
        super(EmsQueryJobConfig, self).__init__(*args, **kwargs)
        self.__priority = priority
        self.__use_query_cache = use_query_cache

    @property
    def priority(self):
        return self.__priority

    @property
    def use_query_cache(self):
        return self.__use_query_cache
